# factorio_tools - Debugging utilities for Factorio
# Copyright (C) 2020  Hornwitser
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse

from . import desync
from . import multi
from . import parse


def main():
    parser = argparse.ArgumentParser(
        prog="factorio_tools",
        description="Debugging utilities for Factorio",
    )
    parser.set_defaults(func=lambda args: parser.print_help())
    subparsers = parser.add_subparsers(help="Tool to run")

    desync_parser(subparsers)
    dat2json_parser(subparsers)
    multi_parser(subparsers)
    args = parser.parse_args()
    args.func(args)

def dat2json_parser(subparsers):
    parser = subparsers.add_parser(
        'dat2json', help="Convert Factorio dat files to json",
        description=
            "Converts the binary formats used in Factorio to json.  "
            "The format options are only necessary if the format cannot be deduced from filename."
    )

    parser.add_argument('--input', '-i', default='-', type=argparse.FileType('rb'), help="input file to convert")
    parser.add_argument('--input-format', help="format of input file")
    parser.add_argument('--output', '-o', default='-', type=argparse.FileType('w'), help="file to output to")
    parser.set_defaults(func=parse.dat2json)

def desync_parser(subparsers):
    parser = subparsers.add_parser(
        'desync', help="Analyze desync reports",
        description=
            "Automated parsing and diffing of desync reports from Factorio"
    )

    parser.add_argument('path', help="Path to desync report, will be extracted if in the .zip file")
    parser.set_defaults(func=desync.analyze)

def multi_parser(subparsers):
    parser = subparsers.add_parser('multi', help="Handle multiple Factorio clients")
    parser.set_defaults(func=lambda args: parser.print_help())
    subparsers = parser.add_subparsers()

    parser_generate_base = subparsers.add_parser('generate-base', help="Generate write dir to base instances on")
    parser_generate_base.add_argument('--base', default="base", help="Name of write dir to generate")
    parser_generate_base.add_argument('--data', help="Path to Factorio data directory")
    parser_generate_base.set_defaults(func=multi.generate_base)

    parser_generate_instances = subparsers.add_parser('generate-instances', help="Generate instances write dirs")
    parser_generate_instances.add_argument('count', type=int, help="Number of instances to generate")
    parser_generate_instances.add_argument('--base', default="base", help="Write dir to base instances on")
    parser_generate_instances.add_argument('--output', default="", help="Path to put generated instances into")
    parser_generate_instances.add_argument('--prefix', default="instance", help="Prefix to name of instance dirs")
    parser_generate_instances.set_defaults(func=multi.generate_instances)

    parser_spawn = subparsers.add_parser('spawn', help="Spawn single client")
    parser_spawn.add_argument('--path', default="base", help="Path to write dir to spawn from")
    parser_spawn.add_argument('--factorio', help="Path to Factorio executable")
    parser_spawn.add_argument('--args', '-a', help="Additional args to pass to Factorio")
    parser_spawn.set_defaults(func=multi.spawn)

    parser_spawn_multi = subparsers.add_parser('spawn-multi', help="Spawn multiple clients")
    parser_spawn_multi.add_argument('--count', '-c', type=int, default=1, help="Clients to spawn")
    parser_spawn_multi.add_argument('--delay', '-d', type=float, default=2.0, help="Deleay between spawns")
    parser_spawn_multi.add_argument('--rows', '-R', type=int, default=4, help="Positioning rows")
    parser_spawn_multi.add_argument('--cols', '-C', type=int, default=5, help="Positioning columns")
    parser_spawn_multi.add_argument('--instance-dirs', default="", help="Location of instance dirs")
    parser_spawn_multi.add_argument('--prefix', default="instance", help="Prefix to name of instance dirs")
    parser_spawn_multi.add_argument('--factorio', help="Path to Factorio executable")
    parser_spawn_multi.add_argument('--args', '-a', help="Additional args to pass to Factorio")
    parser_spawn_multi.set_defaults(func=multi.spawn_multi)

    parser_click = subparsers.add_parser('click', help="Click coordinate on all clients")
    parser_click.add_argument('x', type=int, help="x coordinate")
    parser_click.add_argument('y', type=int, help="y coordinate")
    parser_click.set_defaults(func=multi.click)

if __name__ == '__main__':
    main()



# factorio_tools - Debugging utilities for Factorio
# Copyright (C) 2020  Hornwitser
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
parse.py - Parser for the binary formats to Factorio
"""

import json
import os
import struct
import sys

from construct import \
    Construct, Container, ListContainer, Prefixed, PrefixedArray, Struct, \
    Sequence, Switch, Computed, Terminated, Int32ul, Int24ul, Int16ul, Int8ul, \
    Float64l, Float32l, PascalString, stream_read, stream_write, this


def singleton(cls):
    return cls()

@singleton
class FactorioInt32ul(Construct):
    def _parse(self, stream, context, path):
        tiny = stream_read(stream, 1, path)[0]
        if tiny != 0xff:
            return tiny

        big = stream_read(stream, 4, path)
        return struct.unpack('<l', big)[0]

    def _build(self, obj, stream, context, path):
        if obj < 0xff:
            stream_write(stream, bytes([obj]), 1, path)

        else:
            stream_write(stream, bytes([0xff]), 1, path)
            stream_write(stream, struct.pack('<l', obj), 4, path)

        return obj


    def _sizeof(self, context, path):
        raise SizeofError


object_types = {}

SerializedObject = Struct(
    "type" / Int8ul,
    "data" / Switch(this.type, object_types)
)

# boolean
object_types[0x01] = Struct(
    "uno_" / Int8ul,
    "value" / Int8ul,
)

# float double
object_types[0x02] = Struct(
    "what_" / Int8ul,
    "value" / Float64l,
)

# string
object_types[0x03] = Struct(
    "red_" / Int8ul,
    "blue_" / Int8ul,
    "value" / PascalString(Int8ul, "latin-1"),
)

KeyValuePair = Struct(
    "bloom_" / Int8ul,
    "key" / PascalString(Int8ul, "latin-1"),
    "value" / SerializedObject,
)

# mapping
object_types[0x05] = Struct(
    "monkey_" / Int8ul, # Usually 0
    "items" / PrefixedArray(Int32ul, KeyValuePair),
)

Version = Struct(
    "major" / Int16ul,
    "minor" / Int16ul,
    "patch" / Int16ul,
    "even_" / Int16ul, # Sometimes 1
    "odd_" / Int8ul, # Usually 0
)

OldScriptDat = Struct(
    "_type" / Computed(lambda this: "script"),
    "version" / Version,
    "data" / PrefixedArray(Int32ul, Struct(
        "name" / PascalString(Int8ul, "latin-1"),
        "dump" / PascalString(FactorioInt32ul, "latin-1"),
        "tabletop_" / Int8ul
    )),
    Terminated,
)


script_object_types = {}

ScriptSerializedObject = Struct(
    "type" / Int8ul,
    "data" / Switch(this.type, script_object_types),
)

# boolean
script_object_types[0x02] = Computed(lambda this: True)

# float double
script_object_types[0x03] = Float64l

# string
script_object_types[0x04] = PascalString(FactorioInt32ul, "latin-1")

ScriptKeyValuePair = Struct(
    "key" / ScriptSerializedObject,
    "value" / ScriptSerializedObject,
)

# mapping
script_object_types[0x05] = PrefixedArray(FactorioInt32ul, ScriptKeyValuePair)

# some kind of reference
script_object_types[0x06] = Int16ul

ScriptDat = Struct(
    "_type" / Computed(lambda this: "script"),
    "version" / Version,
    "data" / PrefixedArray(Int32ul, Struct(
        "name" / PascalString(Int8ul, "latin-1"),
        "dump" / Prefixed(FactorioInt32ul, Struct(
            "version" / Version,
            "data" / ScriptSerializedObject,
        )),
        "tabletop_" / Int8ul
    )),
    Terminated,
)

ModSettingsDat = Struct(
    "_type" / Computed(lambda this: "mod-settings"),
    "version" / Version,
    "settings" / SerializedObject,
    Terminated,
)

IdEntry = Struct(
    "type" / PascalString(Int8ul, "latin-1"),
    "names" / PrefixedArray(Int16ul, Struct(
        "name" / PascalString(Int8ul, "latin-1"),
        "id" / Int16ul,
    )),
)

TileIdEntry = Struct(
    "type" / PascalString(Int8ul, "latin-1"),
    "names" / PrefixedArray(Int8ul, Struct(
        "name" / PascalString(Int8ul, "latin-1"),
        "id" / Int8ul,
    )),
)

achievement_types = {
    'achievement': Computed(lambda this: "no-data"),
    'build-entity-achievement': Int32ul,
    'combat-robot-count': Int32ul,
    'construct-with-robots-achievement': Sequence(Int32ul, Int32ul),
    'deconstruct-with-robots-achievement': Int32ul,
    'deliver-by-robots-achievement': Float64l,
    'dont-build-entity-achievement': Int32ul,
    #'dont-craft-manually-achievement': Unknown
    'dont-use-entity-in-energy-production-achievement': Float64l,
    'finish-the-game-achievement': Int32ul,
    'group-attack-achievement': Int32ul,
    'kill-achievement': Float64l,
    'player-damaged-achievement': Sequence(Float32l, Int8ul),
    'produce-achievement': Float64l,
    'produce-per-hour-achievement': Float64l,
    'research-achievement': Computed(lambda this: "no-data"),
    'train-path-achievement': Float64l,
}

def id_to_type(id, table):
    for type_entry in table:
        for name_entry in type_entry.names:
            if id == name_entry.id:
                return name_entry.name, type_entry.type

    raise ValueError(f"No entry in id_table for id {id}")

AchievementEntry = Struct(
    "id" / Int16ul,
    "_name" / Computed(lambda this: id_to_type(this.id, this._._.id_table)[0]),
    "_type" / Computed(lambda this: id_to_type(this.id, this._._.id_table)[1]),
    "data" / Switch(this._type, achievement_types),
)

AchievementDat = Struct(
    "_type" / Computed(lambda this: "achievements"),
    "version" / Version,
    "id_table" / PrefixedArray(Int16ul, IdEntry),
    "achievements" / PrefixedArray(Int32ul, AchievementEntry),
    Terminated,
)

AchievementModdedEntry = Struct(
    "type" / PascalString(Int8ul, "latin-1"),
    "name" / PascalString(Int8ul, "latin-1"),
    "data" / Switch(this.type, achievement_types),
)

AchievementsModdedDat = Struct(
    "_type" / Computed(lambda this: "achievements-modded"),
    "version" / Version,
    "id_table" / PrefixedArray(Int16ul, IdEntry),
    "achievements" / PrefixedArray(Int32ul, AchievementModdedEntry),
    "gremling_" / Int32ul,
    "mole_" / Int16ul,
    Terminated,
)

MigrationEntry = Struct(
    "mod" / PascalString(Int8ul, "latin-1"),
    "migration" / PascalString(Int8ul, "latin-1"),
)

blueprint_types = {}

Blueprint = Struct(
    "bp_type" / Int16ul,
    "bp_name" / PascalString(Int8ul, "latin-1"),
    "content" / Switch(this.bp_type, blueprint_types)
)

BlueprintEntity = Struct(
    "id" / Int16ul,
)

BlueprintData = Struct(
    "version" / Version,
    "applied_migrations" / PrefixedArray(Int8ul, MigrationEntry),
    "entites" / PrefixedArray(Int32ul, BlueprintEntity),

)

# XXX: Hard coded for now
# Blueprint
blueprint_types[0x44] = Struct(
    "rock_" / Int16ul,
    "data" / Prefixed(FactorioInt32ul, BlueprintData),
)
# Blueprint Book
blueprint_types[0x48] = Struct(
    "pages" / PrefixedArray(Int32ul, Struct(
        "pos" / Int32ul,
        "content" / Blueprint,
    )),
)

BlueprintEntry = Struct(
    "ohio_" / PrefixedArray(Int8ul, Int8ul),
    "florida_" / Int8ul,
    "utah_" / Int16ul,
    "brooklyn_" / PrefixedArray(Int8ul, Int16ul),
    "hollywood_" / Int16ul,
    "blueprint" / Blueprint,
)

IdTable = Struct(
    "version" / Version,
    "item_table" / PrefixedArray(Int16ul, IdEntry),
    "tile_table" / PrefixedArray(Int8ul, TileIdEntry),
    "entity_table" / PrefixedArray(Int16ul, IdEntry),
    "recipe_table" / PrefixedArray(Int16ul, IdEntry),
    "fluid_table" / PrefixedArray(Int16ul, IdEntry),
    "signal_table" / PrefixedArray(Int16ul, IdEntry),
    Terminated,
)

from construct import GreedyRange

BlueprintStorageDat = Struct(
    "_type" / Computed(lambda this: "blueprint-storage"),
    "version" / Version,
    "applied_migrations" / PrefixedArray(Int8ul, MigrationEntry),
    "id_table" / Prefixed(FactorioInt32ul, IdTable),

    "chicago_" / Int16ul,
    "jersey_" / Int32ul,
    "buffalo_" / Int32ul,
    "blueprints" / GreedyRange(BlueprintEntry),
)

dat_formats = {
    'achievements': AchievementDat,
    'achievements-modded': AchievementsModdedDat,
    'blueprint-storage': BlueprintStorageDat,
    'mod-settings': ModSettingsDat,
    'script': ScriptDat,
# TODO
#    'crop-cache',
#    'level',
#    'level-init',
}

_excluded_keys = ('_io',)
def container_to_object(obj):
    """Convert a construct container into a json encodable object"""
    if isinstance(obj, Container):
        # Hack for simplyfying script data output
        if (set(obj.keys()) == set(["type", "data", "_io"])):
            if obj.type in (0x02, 0x03, 0x04, 0x05):
                return container_to_object(obj.data)

        return {k: container_to_object(v)
            for k, v in obj.items() if k not in _excluded_keys
        }

    if isinstance(obj, ListContainer):
        return list(map(container_to_object, obj))

    return obj

def dat2json(args):
    if not args.input_format:
        base = os.path.basename(args.input.name)
        if base.endswith('.dat'):
            args.input_format = base[:-4]
        else:
            raise RuntimeError("Unknown input file format")

    Decoder = dat_formats.get(args.input_format)
    if not Decoder:
        raise RuntimeError("Unkown input file format")

    decoded = Decoder.parse_stream(args.input)
    json.dump(container_to_object(decoded), args.output, indent="\t")



# factorio_tools - Debugging utilities for Factorio
# Copyright (C) 2020  Hornwitser
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
multi.py - Spawn and handle multiple Factorio clients
"""

import collections
import configparser
import ctypes
import json
import os
import re
import shutil
import subprocess
import sys
import time


def generate_base(args):
    config_dir = os.path.join(args.base, 'config')
    os.makedirs(config_dir, exist_ok=True)

    if not args.data:
        args.data = os.path.join('__PATH__executable__', '..', '..', 'data')

    config = configparser.ConfigParser()
    config['path'] = {
        'read-data': args.data,
        'write-data': os.path.abspath(args.base),
    }
    config['other'] = { 'check-updates': 'false' }
    config['sound'] = { 'music-volume': '0.000000' }
    config['graphics'] = { 'full-screen': 'false' }

    with open(os.path.join(config_dir, 'config.ini'), 'x', newline='\n') as f:
        f.write('; version=7\n')
        config.write(f, space_around_delimiters=False)

def generate_instances(args):
    with open(os.path.join(args.base, 'player-data.json')) as f:
        player_data = json.load(f)

    base_name = player_data["service-username"]

    for i in range(1, args.count+1):
        dst = os.path.join(args.output, f'{args.prefix}{i}')
        shutil.copytree(args.base, dst, dirs_exist_ok=True)

        player_data["service-username"] = f"{base_name}{i}"
        with open(os.path.join(dst, 'player-data.json'), 'w') as f:
            json.dump(player_data, f, indent=2)

        config_path = os.path.join(dst, 'config', 'config.ini')
        with open(config_path, 'r', newline='\n') as f:
            config_content = f.read()

        def escape(string):
            return string.replace('\\', '\\\\')

        config_content = re.sub(
            r'^write-data=.*$',
            f'write-data={escape(os.path.abspath(dst))}',
            config_content,
            flags=re.MULTILINE
        )

        with open(config_path, 'w', newline='\n') as f:
            f.write(config_content)

def find_factorio():
    """Look for the factorio executable"""
    dir = os.getcwd()
    while True:
        exe = os.path.join(dir, 'bin', 'x64', 'factorio.exe')
        if os.path.exists(exe):
            return exe

        parent = os.path.normpath(os.path.join(dir, '..'))
        if parent == dir:
            break
        dir = parent

    raise RuntimeError(
        "Unable to locate factorio executable, specify it with --factorio"
    )

def spawn(args):
    spawn_one(args.path, args)

def spawn_multi(args):
    instance_index = spawn_instance(1, args)
    for i in range(args.count - 1):
        time.sleep(args.delay)
        instance_index = spawn_instance(instance_index, args)

def spawn_instance(instance_index, args):
    while True:
        write_dir = os.path.join(args.instance_dirs, f'{args.prefix}{instance_index}')
        if not os.path.exists(os.path.join(write_dir, '.lock')):
            break
        instance_index += 1

    pid = spawn_one(write_dir, args)
    move_window(pid, instance_index, args.rows, args.cols)
    return instance_index + 1

def spawn_one(write_dir, args):
    disown = 'pyw -m hornwitser.factorio_tools.disown'
    factorio = args.factorio if args.factorio else find_factorio()
    cfg_path = os.path.join(write_dir, 'config', 'config.ini')
    extra_args = '' if not args.args else f' {args.args}'
    spawn_result = subprocess.run(
        f'{disown} "{factorio} --config {cfg_path}{extra_args}"',
        capture_output=True, text=True
    )
    return int(spawn_result.stdout)


# --- Windows API interactions -----------------------------------------

DWORD = ctypes.c_ulong
BOOL = ctypes.c_long
ULONG = ctypes.c_ulong
LONG = ctypes.c_long
UINT = ctypes.c_uint

user32 = ctypes.windll.user32

class RECT(ctypes.Structure):
    _fields_ = [
        ('left', LONG),
        ('top', LONG),
        ('right', LONG),
        ('bottom', LONG),
    ]

class MONITORINFO(ctypes.Structure):
    _fields_ = [
        ('cbSize', DWORD),
        ('rcMonitor', RECT),
        ('rcWork', RECT),
        ('dwFlags', DWORD),
    ]

MOUSEEVENTF_MOVE        = 0x0001
MOUSEEVENTF_LEFTDOWN    = 0x0002
MOUSEEVENTF_LEFTUP      = 0x0004
MOUSEEVENTF_RIGHTDOWN   = 0x0008
MOUSEEVENTF_RIGHTUP     = 0x0010
MOUSEEVENTF_MIDDLEDOWN  = 0x0020
MOUSEEVENTF_MIDDLEUP    = 0x0040
MOUSEEVENTF_VIRTUALDESK = 0x4000
MOUSEEVENTF_ABSOLUTE    = 0x8000

class MOUSEINPUT(ctypes.Structure):
    _fields_ = [
        ('dx', LONG),
        ('dy', LONG),
        ('mouseData', DWORD),
        ('dwFlags', DWORD),
        ('time', DWORD),
        ('dwExtraInfo', ctypes.POINTER(ULONG)),
    ]


INPUT_MOUSE = 0
INPUT_KEYBOARD = 1
INPUT_HARDWARE = 3

class INPUT_I(ctypes.Union):
    _fields_ = [
        ('mi', MOUSEINPUT),
        # ('ki', KEYBDINPUT),
        # ('hi', HARDWAREINPUT),
    ]

class INPUT(ctypes.Structure):
    _anonymous_ = ("ii",)
    _fields_ = [
        ('type', DWORD),
        ('ii', INPUT_I),
    ]
MONITOR_DEFAULTTOPRIMARY = 0x1

def get_monitor_work_area(hWnd):
    hMonitor = user32.MonitorFromWindow(hWnd, MONITOR_DEFAULTTOPRIMARY)
    info = MONITORINFO(ctypes.sizeof(MONITORINFO))
    success = user32.GetMonitorInfoW(hMonitor, ctypes.byref(info))
    if not success:
        raise RuntimeError("Failed to get monitor info")

    return info.rcWork

# Ported from https://stackoverflow.com/a/21767578
class MainWindowData(ctypes.Structure):
    _fields_ = [
        ('pid', DWORD),
        ('hWnd', DWORD),
    ]

WNDENUMPROC = ctypes.WINFUNCTYPE(BOOL, DWORD, ctypes.POINTER(ctypes.c_void_p))

def find_main_window(pid, tries):
    data = MainWindowData(pid, 0)
    callback = WNDENUMPROC(find_main_window_callback)
    for i in range(tries):
        user32.EnumWindows(callback, ctypes.byref(data))
        if data.hWnd:
            break
        time.sleep(0.1)
    return data.hWnd

def find_main_window_callback(hWnd, p_void):
    p_data = ctypes.cast(p_void, ctypes.POINTER(MainWindowData))
    pid = DWORD()
    user32.GetWindowThreadProcessId(hWnd, ctypes.byref(pid))
    if (pid.value != p_data[0].pid or not is_main_window(hWnd)):
        return True
    p_data[0].hWnd = hWnd
    return False

def is_main_window(hWnd):
    return user32.GetWindow(hWnd, 4) == 0 and user32.IsWindowVisible(hWnd)


FindWindowData = collections.namedtuple('FindWindowData', 'name hWnds')

def find_windows(name):
    """Find Windows who's title starts with name"""
    data = FindWindowData(name, [])
    callback = WNDENUMPROC(find_windows_callback)
    user32.EnumWindows(callback, ctypes.byref(ctypes.py_object(data)))
    return data.hWnds

def find_windows_callback(hWnd, p_void):
    data = ctypes.cast(p_void, ctypes.POINTER(ctypes.py_object))[0]
    if (not is_main_window(hWnd)):
        return True

    buffer = ctypes.create_unicode_buffer(200)
    user32.GetWindowTextW(hWnd, ctypes.byref(buffer), 200)
    name = ctypes.wstring_at(buffer)
    if (name.startswith(data.name)):
        data.hWnds.append(hWnd)

    return True

class POINT(ctypes.Structure):
    _fields_ = [
        ('x', LONG),
        ('y', LONG),
    ]

SW_MAXIMIZE = 3
SW_RESTORE = 9

class WINDOWPLACEMENT(ctypes.Structure):
    _fields_ = [
        ('length', UINT),
        ('flags', UINT),
        ('showCmd', UINT),
        ('ptMinPosition', POINT),
        ('ptMaxPosition', POINT),
        ('rcNormalPosition', RECT),
        ('rcDevice', RECT),
    ]

WM_SYSCOMMAND = 0x0112
SC_RESTORE = 0xf120

def ensure_not_maximized(hWnd):
    window_placement = WINDOWPLACEMENT(ctypes.sizeof(WINDOWPLACEMENT))
    user32.GetWindowPlacement(hWnd, ctypes.byref(window_placement))

    if window_placement.showCmd == SW_MAXIMIZE:
        user32.SendMessageW(hWnd, WM_SYSCOMMAND, SC_RESTORE, 0)

def move_window(pid, instance_index, rows, cols):
    hWnd = find_main_window(pid, 20)
    if not hWnd:
        raise RuntimeError("Unable to find Factorio window")

    ensure_not_maximized(hWnd)
    rcWork = get_monitor_work_area(hWnd)
    width = (rcWork.right - rcWork.left) // cols
    height = (rcWork.bottom - rcWork.top) // rows
    right = rcWork.right - (((instance_index - 1) // rows + 1) * width)
    top = rcWork.top + ((instance_index - 1) % rows * height)

    user32.MoveWindow(hWnd, right, top, width, height, False)

SM_XVIRTUALSCREEN = 76
SM_YVIRTUALSCREEN = 77
SM_CXVIRTUALSCREEN = 78
SM_CYVIRTUALSCREEN = 79

Size = collections.namedtuple('Size', 'x y width height')
virtual_size = None

def get_virtual_size():
    return Size(
        user32.GetSystemMetrics(SM_XVIRTUALSCREEN),
        user32.GetSystemMetrics(SM_YVIRTUALSCREEN),
        user32.GetSystemMetrics(SM_CXVIRTUALSCREEN),
        user32.GetSystemMetrics(SM_CYVIRTUALSCREEN)
    )

def to_virtual(x, y):
    global virtual_size
    if not virtual_size:
        virtual_size = get_virtual_size()

    x = (x - virtual_size.x) * (2**16 - 1) // virtual_size.width
    y = (y - virtual_size.y) * (2**16 - 1) // virtual_size.height
    return x, y

def click_window(hWnd, x, y):
    rect = RECT()
    if not user32.GetWindowRect(hWnd, ctypes.byref(rect)):
        raise RuntimeError("Unable to get size of window")

    flags = (
        MOUSEEVENTF_MOVE
        | MOUSEEVENTF_LEFTDOWN
        | MOUSEEVENTF_LEFTUP
        | MOUSEEVENTF_ABSOLUTE
    )
    x, y = to_virtual(rect.left + x, rect.top + y)
    mi = MOUSEINPUT(x, y, 0, flags, 0, None)
    input = INPUT(INPUT_MOUSE, mi=mi)
    user32.SendInput(1, ctypes.byref(input), ctypes.sizeof(INPUT))

def click(args):
    factorio_windows = find_windows("Factorio")
    for hWnd in factorio_windows:
        click_window(hWnd, args.x, args.y)
        time.sleep(0.1)



# factorio_tools - Debugging utilities for Factorio
# Copyright (C) 2020  Hornwitser
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''
desync.py - Automated analyzis of desync reports
'''

import collections
import difflib
import enum
import itertools
import json
from os import path
import re
import zipfile

from . import parse


BUFFER_SIZE = 2**20
class TokenKind(enum.IntEnum):
    OPEN_TAG = 1
    CLOSE_TAG = 2
    DATA = 3
    COLLAPSED = 4

class Token(collections.namedtuple('Token', 'kind content tag parent pos')):
    def __eq__(self, other):
        return (
            self.kind == other.kind
            and self.content == other.content
            and self.tag == other.tag
        )

    def __ne__(self, other):
        return (
            self.kind != other.kind
            or self.content != other.content
            or self.tag != other.tag
        )

    def __hash__(self):
        return hash((self.kind, self.content, self.tag))


token_spec = [
    (b'OPEN_TAG', br'<(?P<OPEN_TAG_NAME>[a-z-]+)( ([^\x00-\x1f<>\x80-\xff]|<[a-zA-Z]+>)*)?>'),
    (b'CLOSE_TAG', br'</(?P<CLOSE_TAG_NAME>[a-z-]+)>'),
]

tag = re.compile(b'|'.join(b'(?P<%s>%s)' % pair for pair in token_spec))

def tokenize_tagged_file(tagged_file):
    parent = None
    pos = 0
    offset = 0
    buffer = tagged_file.read(BUFFER_SIZE)
    while True:
        match = tag.search(buffer, pos)
        if not match:
            more = tagged_file.read(BUFFER_SIZE)
            if not more:
                yield Token(TokenKind.DATA, buffer[pos:], None, parent, offset + pos)
                break

            buffer += more
            continue

        if match.start() > pos:
            yield Token(TokenKind.DATA, buffer[pos:match.start()], None, parent, offset + pos)

        if match.lastgroup == 'CLOSE_TAG':
            new_parent = parent
            tag_name = match['CLOSE_TAG_NAME']
            while new_parent is not None:
                if new_parent.tag == tag_name:
                    parent = new_parent.parent
                    break

                new_parent = new_parent.parent

            else:
                print(f"Unmatched close tag </{tag_name}>")

        token = Token(TokenKind[match.lastgroup], match[0], match[f'{match.lastgroup}_NAME'], parent, offset + pos)
        yield token
        if match.lastgroup == 'OPEN_TAG':
            parent = token

        pos = match.end()
        if pos > BUFFER_SIZE // 2:
            offset += pos
            buffer = buffer[pos:]
            pos = 0

def token_path(token):
    path = []
    parent = token.parent
    while parent is not None:
        path[0:0] = [parent]
        parent = parent.parent

    formatted = []
    for i, ancestor in enumerate(path):
        formatted.append(f"{'  '*i}{ancestor.content.decode('utf-8')} pos={ancestor.pos}")
    return '\n'.join(formatted)

def collapse(iterator):
    curr = next(iterator, None)
    next1 = next(iterator, None)
    next2 = next(iterator, None)

    def advance():
        nonlocal curr, next1, next2
        curr = next1
        next1 = next2
        next2 = next(iterator, None)

    while curr is not None:
        if (
            curr.kind == TokenKind.OPEN_TAG
            and next1.kind == TokenKind.DATA
            and next2.kind == TokenKind.CLOSE_TAG
            and curr.tag == next2.tag
        ):
            content = curr.content + next1.content + next2.content
            yield Token(TokenKind.COLLAPSED, content, curr.tag, curr.pos)
            advance()
            advance()

        elif (
            curr.kind == TokenKind.OPEN_TAG
            and next1.kind == TokenKind.CLOSE_TAG
            and curr.tag == next1.tag
        ):
            content = curr.content + next1.content
            yield Token(TokenKind.COLLAPSED, content, curr.tag, curr.pos)
            advance()

        else:
            yield curr

        advance()

def find_files(level_zip):
    files = {}
    for name in level_zip.namelist():
        if 'root' not in files:
            files['root'] = name[:name.find('/')]

        if re.match(r'.*/level-heuristic-\d+', name):
            files['heuristic'] = level_zip.open(name)

        if re.match(r'.*/level_with_tags_tick_\d+\.dat', name):
            files['level_with_tags'] = level_zip.open(name)

    files['script'] = level_zip.open(f'{files["root"]}/script.dat')
    return files

def file_differs(a, b):
    try:
        bytes_a = None
        while bytes_a != b'':
            bytes_a = a.read(1024)
            if bytes_a != b.read(1024):
                return True

        return False

    finally:
        a.seek(0)
        b.seek(0)

def diff_script_objects(a, b, path=[]):
    if type(a) is not type(b):
        yield (path, a, b)

    elif type(a) is list:
        for i in range(max(len(a), len(b))):
            sub_path = path + [i] if path != ['data'] else path + [a[i]['name']]
            if i < len(a) and i < len(b):
                yield from diff_script_objects(a[i], b[i], sub_path)
            elif i < len(a):
                yield (sub_path, a[i], None)
            else:
                yield (sub_path, None, b[i])

    elif type(a) is dict:
        shared_keys = {*a.keys(), *b.keys()}
        if shared_keys == {'key', 'value'}:
            path = path[:-1] + [a['key']]
        for k in shared_keys:
            if k in a and k in b:
                yield from diff_script_objects(a[k], b[k], path + [k])
            elif k in a:
                yield (path + [k], a[k], None)
            else:
                yield (path + [k], None, b[k])

    elif a != b:
        yield (path, a, b)


def script_dat_to_object(level_files):
    print(f"parsing {level_files['root']}/script.dat")
    decoded = parse.ScriptDat.parse_stream(level_files['script'])
    level_files['script'].seek(0)
    return parse.container_to_object(decoded)

def diff_tagged_files(a_file, b_file):

    # SequenceMatcher is too slow to work on an entire parsed level
    # so we will instead diff top level tags one by one
    def top_level_tags(generator):
        chunk = []
        for token in generator:
            chunk.append(token)
            if token.parent is None and token.kind != TokenKind.OPEN_TAG:
                yield chunk
                chunk = []

        if chunk:
            yield chunk

    a_generator = top_level_tags(tokenize_tagged_file(a_file))
    b_generator = top_level_tags(tokenize_tagged_file(b_file))

    # For now it's assumed two levels have the same number of chunks
    # and have them in the same order.  This may not actually hold true
    combined = itertools.zip_longest(a_generator, b_generator)
    for a_chunk, b_chunk in combined:
        if a_chunk == b_chunk:
            continue

        if max(len(a_chunk), len(b_chunk)) > 200000:
            print(f"diffing <{a_chunk[0].tag}> {len(a_chunk)}/{len(b_chunk)} tokens, this may take a long time")

        matcher = difflib.SequenceMatcher(None, a_chunk, b_chunk)
        for op, i1, i2, j1, j2 in matcher.get_opcodes():
            if op == 'equal': continue
            print()
            print(f"{op:7}   ref[{i1}:{i2}] -> des[{j1}:{j2}]")
            print(token_path(a_chunk[i1]))
            print(f"ref: {b''.join(map(lambda t: t.content, a_chunk[i1:i2]))!r}")
            print(token_path(b_chunk[j1]))
            print(f"des: {b''.join(map(lambda t: t.content, b_chunk[j1:j2]))}")

def analyze(args):
    if args.path.endswith('.zip'):
        print(f"Unzipping {args.path}")
        report = zipfile.ZipFile(args.path)
        report.extractall(path.dirname(args.path))
        args.path = args.path[:-4]

    ref_zip = zipfile.ZipFile(path.join(args.path, 'reference-level.zip'))
    des_zip = zipfile.ZipFile(path.join(args.path, 'desynced-level.zip'))

    ref_files = find_files(ref_zip)
    des_files = find_files(des_zip)

    if file_differs(ref_files['heuristic'], des_files['heuristic']):
        print()
        print("level-heuristic differs")
        print("-----------------------")
        diff_tagged_files(ref_files['heuristic'], des_files['heuristic'])

    if file_differs(ref_files['script'], des_files['script']):
        print()
        print("script.dat differs")
        print("------------------")

        des_script = script_dat_to_object(des_files)
        ref_script = script_dat_to_object(ref_files)
        for diff in diff_script_objects(ref_script, des_script):
            print(f"Path: {diff[0]}")
            print(f"Reference value: {json.dumps(diff[1])}")
            print(f"Desynced value: {json.dumps(diff[2])}")

    if file_differs(ref_files['level_with_tags'], des_files['level_with_tags']):
        print()
        print("level_with_tags.dat differs")
        print("---------------------------")
        diff_tagged_files(ref_files['level_with_tags'], des_files['level_with_tags'])



# factorio_tools - Debugging utilities for Factorio
# Copyright (C) 2020  Hornwitser
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import setuptools

def read_file(path):
    with open(path) as f:
        return f.read()

setuptools.setup(
    name="hornwitser.factorio_tools",
    version="0.0.2",
    author="Hornwitser",
    author_email="github@hornwitser.no",
    description="Tools for Debugging Factorio",
    long_description=read_file('README.rst'),
    long_description_content_type='text/x-rst',
    url="https://github.com/Hornwitser/factorio_tools",
    namespace_packages=['hornwitser'],
    packages=['hornwitser.factorio_tools'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Environment :: Console",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Operating System :: Microsoft :: Windows",
        "Topic :: Utilities",
    ],
    python_requires='>=3.7',
    install_requires=[
        "construct>=2.10.53",
    ],
    zip_safe=True,
)



# factorio_tools - Debugging utilities for Factorio
# Copyright (C) 2020  Hornwitser
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
disown.py - Dumb Factorio console workaround

Invokes the command specified as the first argument and prints back the
the pid of the command launched over stdout.  This is necessary because
Factorio likes to hijack whatever console it finds in the process that
spawns it.
"""

import sys
import subprocess

if __name__ == '__main__':
    proc = subprocess.Popen(sys.argv[1])
    print(proc.pid, end='')
